@@ -10,18 +10,18 @@ struct _AdwDemoPageAnimations
   AdwAnimation *timed_animation;
   GtkWidget *timed_animation_sample;
   GtkWidget *timed_animation_button_box;
-  GtkSpinButton *timed_animation_repeat_count;
-  GtkSwitch *timed_animation_reverse;
-  GtkSwitch *timed_animation_alternate;
-  GtkSpinButton *timed_animation_duration;
+  AdwSpinRow *timed_animation_repeat_count;
+  AdwSwitchRow *timed_animation_reverse;
+  AdwSwitchRow *timed_animation_alternate;
+  AdwSpinRow *timed_animation_duration;
   AdwComboRow *timed_animation_easing;
   AdwAnimation *spring_animation;
-  GtkSpinButton *spring_animation_velocity;
-  GtkSpinButton *spring_animation_damping;
-  GtkSpinButton *spring_animation_mass;
-  GtkSpinButton *spring_animation_stiffness;
-  GtkSpinButton *spring_animation_epsilon;
-  GtkSwitch *spring_animation_clamp_switch;
+  AdwSpinRow *spring_animation_velocity;
+  AdwSpinRow *spring_animation_damping;
+  AdwSpinRow *spring_animation_mass;
+  AdwSpinRow *spring_animation_stiffness;
+  AdwSpinRow *spring_animation_epsilon;
+  AdwSwitchRow *spring_animation_clamp_switch;
 };
 
 enum {
@@ -33,7 +33,7 @@ enum {
 
 static GParamSpec *props[LAST_PROP];
 
-G_DEFINE_TYPE (AdwDemoPageAnimations, adw_demo_page_animations, ADW_TYPE_BIN)
+G_DEFINE_FINAL_TYPE (AdwDemoPageAnimations, adw_demo_page_animations, ADW_TYPE_BIN)
 
 static AdwAnimation *
 get_current_animation (AdwDemoPageAnimations *self)
@@ -243,9 +243,9 @@ static void
 notify_spring_params_change (AdwDemoPageAnimations *self)
 {
   AdwSpringParams *spring_params =
-    adw_spring_params_new_full (gtk_spin_button_get_value (self->spring_animation_damping),
-                                gtk_spin_button_get_value (self->spring_animation_mass),
-                                gtk_spin_button_get_value (self->spring_animation_stiffness));
+    adw_spring_params_new_full (adw_spin_row_get_value (self->spring_animation_damping),
+                                adw_spin_row_get_value (self->spring_animation_mass),
+                                adw_spin_row_get_value (self->spring_animation_stiffness));
 
   adw_spring_animation_set_spring_params (ADW_SPRING_ANIMATION (self->spring_animation), spring_params);
 
@@ -381,10 +381,10 @@ adw_demo_page_animations_init (AdwDemoPageAnimations *self)
   g_object_bind_property (self->timed_animation_repeat_count, "value",
                           self->timed_animation, "repeat-count",
                           G_BINDING_SYNC_CREATE | G_BINDING_BIDIRECTIONAL);
-  g_object_bind_property (self->timed_animation_reverse, "state",
+  g_object_bind_property (self->timed_animation_reverse, "active",
                           self->timed_animation, "reverse",
                           G_BINDING_SYNC_CREATE | G_BINDING_BIDIRECTIONAL);
-  g_object_bind_property (self->timed_animation_alternate, "state",
+  g_object_bind_property (self->timed_animation_alternate, "active",
                           self->timed_animation, "alternate",
                           G_BINDING_SYNC_CREATE | G_BINDING_BIDIRECTIONAL);
   g_object_bind_property (self->timed_animation_duration, "value",
@@ -407,6 +407,9 @@ adw_demo_page_animations_init (AdwDemoPageAnimations *self)
   adw_timed_animation_set_easing (ADW_TIMED_ANIMATION (self->timed_animation),
                                   ADW_EASE_IN_OUT_CUBIC);
 
+  adw_animation_set_follow_enable_animations_setting (self->timed_animation, FALSE);
+  adw_animation_set_follow_enable_animations_setting (self->spring_animation, FALSE);
+
   g_object_notify_by_pspec (G_OBJECT (self), props[PROP_TIMED_ANIMATION]);
   g_object_notify_by_pspec (G_OBJECT (self), props[PROP_SPRING_ANIMATION]);
 
