@@ -11,9 +11,11 @@
 
 #include "adw-alert-dialog.h"
 
+#include "adw-bin.h"
 #include "adw-dialog-private.h"
 #include "adw-gizmo-private.h"
 #include "adw-gtkbuilder-utils-private.h"
+#include "adw-length-unit.h"
 #include "adw-marshalers.h"
 #include "adw-widget-utils-private.h"
 
@@ -77,8 +79,7 @@
  * ## Async API
  *
  * `AdwAlertDialog` can also be used via the [method@AlertDialog.choose] method.
- * This API follows the GIO async pattern, and the result can be obtained by
- * calling [method@AlertDialog.choose_finish], for example:
+ * This API follows the GIO async pattern, for example:
  *
  * ```c
  * static void
@@ -123,7 +124,7 @@
  *
  * `AdwAlertDialog` supports adding responses in UI definitions by via the
  * `<responses>` element that may contain multiple `<response>` elements, each
- * respresenting a response.
+ * representing a response.
  *
  * Each of the `<response>` elements must have the `id` attribute specifying the
  * response ID. The contents of the element are used as the response label.
@@ -171,8 +172,13 @@
  * Since: 1.2
  */
 
-#define DIALOG_MAX_WIDTH 550
+#define DIALOG_MAX_WIDTH 372 /* sp, not px */
+#define DIALOG_MAX_WIDE_WIDTH 600 /* sp, not px */
+#define DIALOG_PREFERRED_WIDTH 300 /* sp, not px */
 #define DIALOG_MIN_WIDTH 300
+#define BUTTON_SPACING 12
+#define RESPONSE_HORZ_PADDING 48
+#define RESPONSE_HORZ_PADDING_SHORT 36
 
 typedef struct {
   AdwAlertDialog *dialog;
@@ -182,7 +188,6 @@ typedef struct {
   gboolean enabled;
 
   GtkWidget *button;
-  GtkWidget *separator;
 } ResponseInfo;
 
 typedef struct
@@ -190,8 +195,11 @@ typedef struct
   GtkWidget *contents;
   GtkWidget *window_handle;
   GtkWidget *scrolled_window;
+  GtkWidget *heading_bin;
   GtkWidget *heading_label;
+  GtkWidget *heading_label_small;
   GtkWidget *body_label;
+  GtkWidget *child_bin;
   GtkBox *message_area;
   GtkWidget *response_area;
 
@@ -200,6 +208,7 @@ typedef struct
   char *body;
   gboolean body_use_markup;
   GtkWidget *child;
+  gboolean prefer_wide_layout;
 
   GList *responses;
   GHashTable *id_to_response;
@@ -207,6 +216,7 @@ typedef struct
   GQuark close_response;
 
   gboolean block_close_response;
+  gboolean is_short;
 } AdwAlertDialogPrivate;
 
 static void adw_alert_dialog_buildable_init (GtkBuildableIface *iface);
@@ -224,6 +234,7 @@ enum {
   PROP_BODY,
   PROP_BODY_USE_MARKUP,
   PROP_EXTRA_CHILD,
+  PROP_PREFER_WIDE_LAYOUT,
   PROP_DEFAULT_RESPONSE,
   PROP_CLOSE_RESPONSE,
   LAST_PROP,
@@ -253,37 +264,43 @@ find_response (AdwAlertDialog *self,
   return g_hash_table_lookup (priv->id_to_response, id);
 }
 
-static void
-button_clicked_cb (ResponseInfo *info)
+
+static inline void
+emit_response (AdwAlertDialog *self, GQuark response)
 {
-  AdwAlertDialog *self = info->dialog;
   AdwAlertDialogPrivate *priv = adw_alert_dialog_get_instance_private (self);
 
   g_object_ref (self);
   priv->block_close_response = TRUE;
 
   adw_dialog_close (ADW_DIALOG (self));
-  g_signal_emit (self, signals[SIGNAL_RESPONSE], info->id, g_quark_to_string (info->id));
+  g_signal_emit (self, signals[SIGNAL_RESPONSE], response, g_quark_to_string (response));
 
   priv->block_close_response = FALSE;
   g_object_unref (self);
 }
 
+
+static void
+button_clicked_cb (ResponseInfo *info)
+{
+  emit_response (info->dialog, info->id);
+}
+
 static GtkWidget *
 create_response_button (AdwAlertDialog *self,
                         ResponseInfo   *info)
 {
   GtkWidget *button = gtk_button_new_with_mnemonic (info->label);
 
-  gtk_widget_add_css_class (button, "flat");
   gtk_button_set_can_shrink (GTK_BUTTON (button), TRUE);
 
   switch (info->appearance) {
   case ADW_RESPONSE_SUGGESTED:
-    gtk_widget_add_css_class (button, "suggested");
+    gtk_widget_add_css_class (button, "suggested-action");
     break;
   case ADW_RESPONSE_DESTRUCTIVE:
-    gtk_widget_add_css_class (button, "destructive");
+    gtk_widget_add_css_class (button, "destructive-action");
     break;
   case ADW_RESPONSE_DEFAULT:
   default:
@@ -341,18 +358,30 @@ static void
 adw_alert_dialog_map (GtkWidget *widget)
 {
   AdwAlertDialog *self = ADW_ALERT_DIALOG (widget);
-  AdwAlertDialogPrivate *priv = adw_alert_dialog_get_instance_private (self);
-  GtkWidget *focus;
   GtkWidget *window;
 
   GTK_WIDGET_CLASS (adw_alert_dialog_parent_class)->map (widget);
 
   window = adw_dialog_get_window (ADW_DIALOG (self));
 
-  if (window)
+  if (window) {
     gtk_widget_add_css_class (window, "alert");
+    gtk_widget_grab_focus (widget);
+  }
+}
+
+static gboolean
+adw_alert_dialog_grab_focus (GtkWidget *widget)
+{
+  AdwAlertDialog *self = ADW_ALERT_DIALOG (widget);
+  AdwAlertDialogPrivate *priv = adw_alert_dialog_get_instance_private (self);
+  GtkWidget *focus;
+  gboolean ret;
 
   focus = adw_dialog_get_focus (ADW_DIALOG (self));
+
+  ret = GTK_WIDGET_CLASS (adw_alert_dialog_parent_class)->grab_focus (widget);
+
   if (!focus) {
     GtkWidget *default_widget;
     GList *l;
@@ -363,14 +392,12 @@ adw_alert_dialog_map (GtkWidget *widget)
       if (GTK_IS_LABEL (focus) && !gtk_label_get_current_uri (GTK_LABEL (focus)))
         gtk_label_select_region (GTK_LABEL (focus), 0, 0);
 
-      return;
+      return TRUE;
     }
 
     default_widget = adw_dialog_get_default_widget (ADW_DIALOG (self));
-    if (default_widget) {
-      gtk_widget_grab_focus (default_widget);
-      return;
-    }
+    if (default_widget)
+      return gtk_widget_grab_focus (default_widget);
 
     for (l = g_list_last (priv->responses); l; l = l->prev) {
       ResponseInfo *response = l->data;
@@ -378,10 +405,110 @@ adw_alert_dialog_map (GtkWidget *widget)
       if (!response->enabled)
         continue;
 
-      gtk_widget_grab_focus (response->button);
-      return;
+      return gtk_widget_grab_focus (response->button);
     }
   }
+
+  return ret;
+}
+
+static GtkSizeRequestMode
+get_heading_request_mode (GtkWidget *widget)
+{
+  return GTK_SIZE_REQUEST_HEIGHT_FOR_WIDTH;
+}
+
+static void
+measure_heading (GtkWidget      *widget,
+                 GtkOrientation  orientation,
+                 int             for_size,
+                 int            *minimum,
+                 int            *natural,
+                 int            *minimum_baseline,
+                 int            *natural_baseline)
+{
+  AdwAlertDialog *self = g_object_get_data (G_OBJECT (widget), "-adw-alert-dialog");
+  AdwAlertDialogPrivate *priv = adw_alert_dialog_get_instance_private (self);
+  int large_min, large_nat, small_min, small_nat;
+
+  if (priv->is_short) {
+    gtk_widget_measure (priv->heading_label_small, orientation, for_size,
+                        minimum, natural, NULL, NULL);
+
+    if (minimum_baseline)
+      *minimum_baseline = -1;
+    if (natural_baseline)
+      *natural_baseline = -1;
+
+    return;
+  }
+
+  gtk_widget_measure (priv->heading_label, orientation, for_size,
+                      &large_min, &large_nat, NULL, NULL);
+  gtk_widget_measure (priv->heading_label_small, orientation, for_size,
+                      &small_min, &small_nat, NULL, NULL);
+
+  if (orientation == GTK_ORIENTATION_HORIZONTAL) {
+    if (minimum)
+      *minimum = MIN (large_min, small_min);
+    if (natural)
+      *natural = MAX (large_nat, small_nat);
+  } else {
+    if (for_size < 0) {
+      if (minimum)
+        *minimum = MAX (large_min, small_min);
+      if (natural)
+        *natural = MAX (large_nat, small_nat);
+    } else {
+      int large_width;
+
+      gtk_widget_measure (priv->heading_label, GTK_ORIENTATION_HORIZONTAL, -1,
+                          NULL, &large_width, NULL, NULL);
+
+      if (minimum)
+        *minimum = (large_width > for_size) ? small_min : large_min;
+      if (natural)
+        *natural = (large_width > for_size) ? small_nat : large_nat;
+    }
+  }
+
+  if (minimum_baseline)
+    *minimum_baseline = -1;
+  if (natural_baseline)
+    *natural_baseline = -1;
+}
+
+static void
+allocate_heading (GtkWidget *widget,
+                  int        width,
+                  int        height,
+                  int        baseline)
+{
+  AdwAlertDialog *self = g_object_get_data (G_OBJECT (widget), "-adw-alert-dialog");
+  AdwAlertDialogPrivate *priv = adw_alert_dialog_get_instance_private (self);
+  gboolean small_label;
+
+  if (priv->is_short) {
+    small_label = TRUE;
+  } else {
+    int large_nat;
+
+    gtk_widget_measure (priv->heading_label, GTK_ORIENTATION_HORIZONTAL, -1,
+                        NULL, &large_nat, NULL, NULL);
+
+    small_label = large_nat > width;
+  }
+
+  if (gtk_widget_get_child_visible (priv->heading_label) == small_label)
+    gtk_widget_set_child_visible (priv->heading_label, !small_label);
+
+  if (gtk_widget_get_child_visible (priv->heading_label_small) != small_label)
+    gtk_widget_set_child_visible (priv->heading_label_small, small_label);
+
+  if (small_label)
+    gtk_widget_allocate (priv->heading_label_small, width, height, baseline, NULL);
+  else
+    gtk_widget_allocate (priv->heading_label, width, height, baseline, NULL);
 }
 
 static void
@@ -417,17 +544,9 @@ measure_responses_do (AdwAlertDialog *self,
       nat += child_nat;
     }
 
-    if (response->separator) {
-      gtk_widget_measure (response->separator, orientation, -1,
-                          &child_min, &child_nat, NULL, NULL);
-
-      if (horiz == compact) {
-        min = MAX (min, child_min);
-        nat = MAX (nat, child_nat);
-      } else {
-        min += child_min;
-        nat += child_nat;
-      }
+    if (horiz != compact && l->next) {
+      min += BUTTON_SPACING;
+      nat += BUTTON_SPACING;
     }
   }
 
@@ -457,19 +576,21 @@ measure_responses (GtkWidget      *widget,
                    int            *minimum_baseline,
                    int            *natural_baseline)
 {
-  AdwAlertDialog *self = ADW_ALERT_DIALOG (gtk_widget_get_ancestor (widget, ADW_TYPE_ALERT_DIALOG));
+  AdwAlertDialog *self = g_object_get_data (G_OBJECT (widget), "-adw-alert-dialog");
+  AdwAlertDialogPrivate *priv = adw_alert_dialog_get_instance_private (self);
 
   if (orientation == GTK_ORIENTATION_HORIZONTAL) {
     measure_responses_do (self, TRUE, orientation, minimum, NULL);
     measure_responses_do (self, FALSE, orientation, NULL, natural);
   } else {
     int wide_nat = 0;
+    gboolean use_compact_layout;
 
     if (for_size >= 0)
       measure_responses_do (self, FALSE, GTK_ORIENTATION_HORIZONTAL, NULL, &wide_nat);
 
-    measure_responses_do (self, for_size >= 0 && for_size < wide_nat,
-                          orientation, minimum, natural);
+    use_compact_layout = for_size >= 0 && wide_nat > for_size && !priv->is_short;
+    measure_responses_do (self, use_compact_layout, orientation, minimum, natural);
   }
 
   if (minimum_baseline)
@@ -491,7 +612,7 @@ allocate_responses (GtkWidget *widget,
 
   measure_responses_do (self, FALSE, GTK_ORIENTATION_HORIZONTAL, NULL, &wide_nat);
 
-  compact = wide_nat > width;
+  compact = wide_nat > width && !priv->is_short;
 
   if (compact)
     gtk_widget_add_css_class (widget, "compact");
@@ -506,16 +627,6 @@ allocate_responses (GtkWidget *widget,
       ResponseInfo *response = l->data;
       int child_height;
 
-      if (response->separator) {
-        gtk_widget_measure (response->separator, GTK_ORIENTATION_VERTICAL, -1,
-                            &child_height, NULL, NULL, NULL);
-
-        pos -= child_height;
-
-        gtk_widget_allocate (response->separator, width, child_height, -1,
-                             gsk_transform_translate (NULL, &GRAPHENE_POINT_INIT (0, pos)));
-      }
-
       gtk_widget_measure (response->button, GTK_ORIENTATION_VERTICAL, -1,
                           &child_height, NULL, NULL, NULL);
 
@@ -523,48 +634,19 @@ allocate_responses (GtkWidget *widget,
 
       gtk_widget_allocate (response->button, width, child_height, -1,
                            gsk_transform_translate (NULL, &GRAPHENE_POINT_INIT (0, pos)));
+
+      pos -= BUTTON_SPACING;
     }
   } else {
     gboolean is_rtl = gtk_widget_get_direction (widget) == GTK_TEXT_DIR_RTL;
     int pos = is_rtl ? width : 0;
     int n_buttons = g_list_length (priv->responses);
-    int total_width = width;
-    int button_width;
+    int total_width = width - BUTTON_SPACING * MAX (0, (n_buttons - 1));
+    int button_width = (int) ceil ((double) total_width / n_buttons);
     GList *l;
 
     for (l = priv->responses; l; l = l->next) {
       ResponseInfo *response = l->data;
-      int separator_width;
-
-      if (!response->separator)
-        continue;
-
-      gtk_widget_measure (response->separator, GTK_ORIENTATION_HORIZONTAL, -1,
-                          &separator_width, NULL, NULL, NULL);
-
-      total_width -= separator_width;
-    }
-
-    button_width = (int) ceil ((double) total_width / n_buttons);
-
-    for (l = priv->responses; l; l = l->next) {
-      ResponseInfo *response = l->data;
-
-      if (response->separator) {
-        int separator_width;
-
-        gtk_widget_measure (response->separator, GTK_ORIENTATION_HORIZONTAL, -1,
-                            &separator_width, NULL, NULL, NULL);
-
-        if (is_rtl)
-          pos -= separator_width;
-
-        gtk_widget_allocate (response->separator, separator_width, height, -1,
-                             gsk_transform_translate (NULL, &GRAPHENE_POINT_INIT (pos, 0)));
-
-        if (!is_rtl)
-          pos += separator_width;
-      }
 
       button_width = MIN (button_width, total_width);
 
@@ -577,7 +659,9 @@ allocate_responses (GtkWidget *widget,
                            gsk_transform_translate (NULL, &GRAPHENE_POINT_INIT (pos, 0)));
 
       if (!is_rtl)
-        pos += button_width;
+        pos += button_width + BUTTON_SPACING;
+      else
+        pos -= BUTTON_SPACING;
     }
   }
 }
@@ -591,7 +675,7 @@ measure_child (GtkWidget      *widget,
                int            *min_baseline,
                int            *nat_baseline)
 {
-  AdwAlertDialog *self = ADW_ALERT_DIALOG (gtk_widget_get_ancestor (widget, ADW_TYPE_ALERT_DIALOG));
+  AdwAlertDialog *self = g_object_get_data (G_OBJECT (widget), "-adw-alert-dialog");
   AdwAlertDialogPrivate *priv = adw_alert_dialog_get_instance_private (self);
   int max_size, min_size, base_nat;
 
@@ -599,18 +683,51 @@ measure_child (GtkWidget      *widget,
                       &min_size, &base_nat, NULL, NULL);
 
   if (orientation == GTK_ORIENTATION_HORIZONTAL) {
-    int wide_nat, narrow_nat;
+    int pref_width, wide_nat, narrow_nat, heading_nat;
+    int regular_max_size;
 
     min_size = MAX (min_size, DIALOG_MIN_WIDTH);
-    max_size = DIALOG_MAX_WIDTH;
 
-    measure_responses_do (self, FALSE, GTK_ORIENTATION_HORIZONTAL, NULL, &wide_nat);
-    measure_responses_do (self, TRUE, GTK_ORIENTATION_HORIZONTAL, NULL, &narrow_nat);
+    pref_width = adw_length_unit_to_px (ADW_LENGTH_UNIT_SP,
+                                        DIALOG_PREFERRED_WIDTH,
+                                        gtk_widget_get_settings (widget));
 
-    narrow_nat = MAX (narrow_nat, DIALOG_MIN_WIDTH);
+    if (gtk_widget_get_visible (priv->heading_bin)) {
+      gtk_widget_measure (priv->heading_bin, orientation, -1,
+                          NULL, &heading_nat, NULL, NULL);
+    } else {
+      heading_nat = 0;
+    }
 
-    if (max_size < wide_nat)
-      max_size = MIN (max_size, narrow_nat);
+    measure_responses_do (self, FALSE, orientation, NULL, &wide_nat);
+    measure_responses_do (self, TRUE, orientation, NULL, &narrow_nat);
+
+    if (priv->is_short) {
+      wide_nat += RESPONSE_HORZ_PADDING_SHORT;
+      narrow_nat += RESPONSE_HORZ_PADDING_SHORT;
+    } else {
+      wide_nat += RESPONSE_HORZ_PADDING;
+      narrow_nat += RESPONSE_HORZ_PADDING;
+    }
+
+    narrow_nat = MAX (narrow_nat, pref_width);
+    heading_nat = CLAMP (heading_nat, narrow_nat, wide_nat);
+
+    regular_max_size = adw_length_unit_to_px (ADW_LENGTH_UNIT_SP,
+                                              DIALOG_MAX_WIDTH,
+                                              gtk_widget_get_settings (widget));
+
+    if (priv->prefer_wide_layout || priv->is_short) {
+      max_size = adw_length_unit_to_px (ADW_LENGTH_UNIT_SP,
+                                        DIALOG_MAX_WIDE_WIDTH,
+                                        gtk_widget_get_settings (widget));
+
+      max_size = MIN (max_size, wide_nat);
+    } else if (wide_nat > min_size) {
+      max_size = MIN (regular_max_size, heading_nat);
+    } else {
+      max_size = regular_max_size;
+    }
   } else {
     max_size = G_MAXINT;
   }
@@ -633,12 +750,26 @@ allocate_child (GtkWidget *widget,
                 int        height,
                 int        baseline)
 {
-  AdwAlertDialog *self = ADW_ALERT_DIALOG (gtk_widget_get_ancestor (widget, ADW_TYPE_ALERT_DIALOG));
+  AdwAlertDialog *self = g_object_get_data (G_OBJECT (widget), "-adw-alert-dialog");
   AdwAlertDialogPrivate *priv = adw_alert_dialog_get_instance_private (self);
 
   gtk_widget_allocate (priv->window_handle, width, height, baseline, NULL);
 }
 
+static void
+css_classes_changed_cb (AdwAlertDialog *self)
+{
+  AdwAlertDialogPrivate *priv = adw_alert_dialog_get_instance_private (self);
+  gboolean is_short = gtk_widget_has_css_class (GTK_WIDGET (self), "short");
+
+  if (is_short == priv->is_short)
+    return;
+
+  priv->is_short = is_short;
+
+  gtk_widget_queue_resize (priv->contents);
+}
+
 static void
 adw_alert_dialog_get_property (GObject    *object,
                                guint       prop_id,
@@ -663,6 +794,9 @@ adw_alert_dialog_get_property (GObject    *object,
   case PROP_EXTRA_CHILD:
     g_value_set_object (value, adw_alert_dialog_get_extra_child (self));
     break;
+  case PROP_PREFER_WIDE_LAYOUT:
+    g_value_set_boolean (value, adw_alert_dialog_get_prefer_wide_layout (self));
+    break;
   case PROP_DEFAULT_RESPONSE:
     g_value_set_string (value, adw_alert_dialog_get_default_response (self));
     break;
@@ -698,6 +832,9 @@ adw_alert_dialog_set_property (GObject      *object,
   case PROP_EXTRA_CHILD:
     adw_alert_dialog_set_extra_child (self, g_value_get_object (value));
     break;
+  case PROP_PREFER_WIDE_LAYOUT:
+    adw_alert_dialog_set_prefer_wide_layout (self, g_value_get_boolean (value));
+    break;
   case PROP_DEFAULT_RESPONSE:
     adw_alert_dialog_set_default_response (self, g_value_get_string (value));
     break;
@@ -752,11 +889,12 @@ adw_alert_dialog_class_init (AdwAlertDialogClass *klass)
   object_class->finalize = adw_alert_dialog_finalize;
 
   widget_class->map = adw_alert_dialog_map;
+  widget_class->grab_focus = adw_alert_dialog_grab_focus;
 
   dialog_class->closed = adw_alert_dialog_closed;
 
   /**
-   * AdwAlertDialog:heading: (attributes org.gtk.Property.get=adw_alert_dialog_get_heading org.gtk.Property.set=adw_alert_dialog_set_heading)
+   * AdwAlertDialog:heading:
    *
    * The heading of the dialog.
    *
@@ -768,7 +906,7 @@ adw_alert_dialog_class_init (AdwAlertDialogClass *klass)
                          G_PARAM_READWRITE | G_PARAM_STATIC_STRINGS | G_PARAM_EXPLICIT_NOTIFY);
 
   /**
-   * AdwAlertDialog:heading-use-markup: (attributes org.gtk.Property.get=adw_alert_dialog_get_heading_use_markup org.gtk.Property.set=adw_alert_dialog_set_heading_use_markup)
+   * AdwAlertDialog:heading-use-markup:
    *
    * Whether the heading includes Pango markup.
    *
@@ -782,7 +920,7 @@ adw_alert_dialog_class_init (AdwAlertDialogClass *klass)
                           G_PARAM_READWRITE | G_PARAM_STATIC_STRINGS | G_PARAM_EXPLICIT_NOTIFY);
 
   /**
-   * AdwAlertDialog:body: (attributes org.gtk.Property.get=adw_alert_dialog_get_body org.gtk.Property.set=adw_alert_dialog_set_body)
+   * AdwAlertDialog:body:
    *
    * The body text of the dialog.
    *
@@ -794,7 +932,7 @@ adw_alert_dialog_class_init (AdwAlertDialogClass *klass)
                          G_PARAM_READWRITE | G_PARAM_STATIC_STRINGS | G_PARAM_EXPLICIT_NOTIFY);
 
   /**
-   * AdwAlertDialog:body-use-markup: (attributes org.gtk.Property.get=adw_alert_dialog_get_body_use_markup org.gtk.Property.set=adw_alert_dialog_set_body_use_markup)
+   * AdwAlertDialog:body-use-markup:
    *
    * Whether the body text includes Pango markup.
    *
@@ -808,7 +946,7 @@ adw_alert_dialog_class_init (AdwAlertDialogClass *klass)
                           G_PARAM_READWRITE | G_PARAM_STATIC_STRINGS | G_PARAM_EXPLICIT_NOTIFY);
 
   /**
-   * AdwAlertDialog:extra-child: (attributes org.gtk.Property.get=adw_alert_dialog_get_extra_child org.gtk.Property.set=adw_alert_dialog_set_extra_child)
+   * AdwAlertDialog:extra-child:
    *
    * The child widget.
    *
@@ -822,7 +960,22 @@ adw_alert_dialog_class_init (AdwAlertDialogClass *klass)
                          G_PARAM_READWRITE | G_PARAM_STATIC_STRINGS | G_PARAM_EXPLICIT_NOTIFY);
 
   /**
-   * AdwAlertDialog:default-response: (attributes org.gtk.Property.get=adw_alert_dialog_get_default_response org.gtk.Property.set=adw_alert_dialog_set_default_response)
+   * AdwAlertDialog:prefer-wide-layout:
+   *
+   * Whether to prefer wide layout.
+   *
+   * Prefer horizontal button layout when possible, and wider dialog width
+   * otherwise.
+   *
+   * Since: 1.6
+   */
+  props[PROP_PREFER_WIDE_LAYOUT] =
+    g_param_spec_boolean ("prefer-wide-layout", NULL, NULL,
+                          FALSE,
+                          G_PARAM_READWRITE | G_PARAM_STATIC_STRINGS | G_PARAM_EXPLICIT_NOTIFY);
+
+  /**
+   * AdwAlertDialog:default-response:
    *
    * The response ID of the default response.
    *
@@ -839,7 +992,7 @@ adw_alert_dialog_class_init (AdwAlertDialogClass *klass)
                           G_PARAM_READWRITE | G_PARAM_STATIC_STRINGS | G_PARAM_EXPLICIT_NOTIFY);
 
   /**
-   * AdwAlertDialog:close-response: (attributes org.gtk.Property.get=adw_alert_dialog_get_close_response org.gtk.Property.set=adw_alert_dialog_set_close_response)
+   * AdwAlertDialog:close-response:
    *
    * The ID of the close response.
    *
@@ -895,8 +1048,11 @@ adw_alert_dialog_class_init (AdwAlertDialogClass *klass)
   gtk_widget_class_bind_template_child_private (widget_class, AdwAlertDialog, contents);
   gtk_widget_class_bind_template_child_private (widget_class, AdwAlertDialog, window_handle);
   gtk_widget_class_bind_template_child_private (widget_class, AdwAlertDialog, scrolled_window);
+  gtk_widget_class_bind_template_child_private (widget_class, AdwAlertDialog, heading_bin);
   gtk_widget_class_bind_template_child_private (widget_class, AdwAlertDialog, heading_label);
+  gtk_widget_class_bind_template_child_private (widget_class, AdwAlertDialog, heading_label_small);
   gtk_widget_class_bind_template_child_private (widget_class, AdwAlertDialog, body_label);
+  gtk_widget_class_bind_template_child_private (widget_class, AdwAlertDialog, child_bin);
   gtk_widget_class_bind_template_child_private (widget_class, AdwAlertDialog, message_area);
   gtk_widget_class_bind_template_child_private (widget_class, AdwAlertDialog, response_area);
 
@@ -918,11 +1074,20 @@ adw_alert_dialog_init (AdwAlertDialog *self)
 
   gtk_widget_init_template (GTK_WIDGET (self));
 
+  g_object_set_data (G_OBJECT (priv->contents), "-adw-alert-dialog", self);
+  g_object_set_data (G_OBJECT (priv->heading_bin), "-adw-alert-dialog", self);
+  g_object_set_data (G_OBJECT (priv->response_area), "-adw-alert-dialog", self);
+
   gtk_widget_set_layout_manager (priv->contents,
                                  gtk_custom_layout_new (adw_widget_get_request_mode,
                                                         measure_child,
                                                         allocate_child));
 
+  gtk_widget_set_layout_manager (priv->heading_bin,
+                                 gtk_custom_layout_new (get_heading_request_mode,
+                                                        measure_heading,
+                                                        allocate_heading));
+
   gtk_widget_set_layout_manager (priv->response_area,
                                  gtk_custom_layout_new (get_responses_request_mode,
                                                         measure_responses,
@@ -933,10 +1098,20 @@ adw_alert_dialog_init (AdwAlertDialog *self)
   adw_gizmo_set_grab_focus_func (ADW_GIZMO (priv->contents),
                             (AdwGizmoGrabFocusFunc) adw_widget_grab_focus_child);
 
+  adw_gizmo_set_focus_func (ADW_GIZMO (priv->heading_bin),
+                            (AdwGizmoFocusFunc) adw_widget_focus_child);
+  adw_gizmo_set_grab_focus_func (ADW_GIZMO (priv->heading_bin),
+                            (AdwGizmoGrabFocusFunc) adw_widget_grab_focus_child);
+
   adw_gizmo_set_focus_func (ADW_GIZMO (priv->response_area),
                             (AdwGizmoFocusFunc) adw_widget_focus_child);
   adw_gizmo_set_grab_focus_func (ADW_GIZMO (priv->response_area),
                             (AdwGizmoGrabFocusFunc) adw_widget_grab_focus_child);
+
+  g_signal_connect (self, "notify::css-classes",
+                    G_CALLBACK (css_classes_changed_cb), self);
+
+  css_classes_changed_cb (self);
 }
 
 /* Custom tag handling was copied and modified
@@ -1210,7 +1385,7 @@ adw_alert_dialog_new (const char *heading,
 }
 
 /**
- * adw_alert_dialog_get_heading: (attributes org.gtk.Method.get_property=heading)
+ * adw_alert_dialog_get_heading:
  * @self: an alert dialog
  *
  * Gets the heading of @self.
@@ -1232,7 +1407,7 @@ adw_alert_dialog_get_heading (AdwAlertDialog *self)
 }
 
 /**
- * adw_alert_dialog_set_heading: (attributes org.gtk.Method.set_property=heading)
+ * adw_alert_dialog_set_heading:
  * @self: an alert dialog
  * @heading: (nullable): the heading of @self
  *
@@ -1255,7 +1430,9 @@ adw_alert_dialog_set_heading (AdwAlertDialog *self,
     return;
 
   gtk_label_set_label (GTK_LABEL (priv->heading_label), heading);
-  gtk_widget_set_visible (priv->heading_label, heading && *heading);
+  gtk_label_set_label (GTK_LABEL (priv->heading_label_small), heading);
+
+  gtk_widget_set_visible (priv->heading_bin, heading && *heading);
 
   if (heading && *heading)
     gtk_widget_add_css_class (GTK_WIDGET (priv->message_area), "has-heading");
@@ -1268,7 +1445,7 @@ adw_alert_dialog_set_heading (AdwAlertDialog *self,
 }
 
 /**
- * adw_alert_dialog_get_heading_use_markup: (attributes org.gtk.Method.get_property=heading-use-markup)
+ * adw_alert_dialog_get_heading_use_markup:
  * @self: an alert dialog
  *
  * Gets whether the heading of @self includes Pango markup.
@@ -1290,7 +1467,7 @@ adw_alert_dialog_get_heading_use_markup (AdwAlertDialog *self)
 }
 
 /**
- * adw_alert_dialog_set_heading_use_markup: (attributes org.gtk.Method.set_property=heading-use-markup)
+ * adw_alert_dialog_set_heading_use_markup:
  * @self: an alert dialog
  * @use_markup: whether to use markup for heading
  *
@@ -1318,6 +1495,7 @@ adw_alert_dialog_set_heading_use_markup (AdwAlertDialog *self,
   priv->heading_use_markup = use_markup;
 
   gtk_label_set_use_markup (GTK_LABEL (priv->heading_label), use_markup);
+  gtk_label_set_use_markup (GTK_LABEL (priv->heading_label_small), use_markup);
 
   update_window_title (self);
 
@@ -1418,7 +1596,7 @@ adw_alert_dialog_format_heading_markup (AdwAlertDialog *self,
 }
 
 /**
- * adw_alert_dialog_get_body: (attributes org.gtk.Method.get_property=body)
+ * adw_alert_dialog_get_body:
  * @self: an alert dialog
  *
  * Gets the body text of @self.
@@ -1440,7 +1618,7 @@ adw_alert_dialog_get_body (AdwAlertDialog *self)
 }
 
 /**
- * adw_alert_dialog_set_body: (attributes org.gtk.Method.set_property=body)
+ * adw_alert_dialog_set_body:
  * @self: an alert dialog
  * @body: the body of @self
  *
@@ -1474,7 +1652,7 @@ adw_alert_dialog_set_body (AdwAlertDialog *self,
 }
 
 /**
- * adw_alert_dialog_get_body_use_markup: (attributes org.gtk.Method.get_property=body-use-markup)
+ * adw_alert_dialog_get_body_use_markup:
  * @self: an alert dialog
  *
  * Gets whether the body text of @self includes Pango markup.
@@ -1496,7 +1674,7 @@ adw_alert_dialog_get_body_use_markup (AdwAlertDialog *self)
 }
 
 /**
- * adw_alert_dialog_set_body_use_markup: (attributes org.gtk.Method.set_property=body-use-markup)
+ * adw_alert_dialog_set_body_use_markup:
  * @self: an alert dialog
  * @use_markup: whether to use markup for body text
  *
@@ -1621,7 +1799,7 @@ adw_alert_dialog_format_body_markup (AdwAlertDialog *self,
 }
 
 /**
- * adw_alert_dialog_get_extra_child: (attributes org.gtk.Method.get_property=extra-child)
+ * adw_alert_dialog_get_extra_child:
  * @self: an alert dialog
  *
  * Gets the child widget of @self.
@@ -1643,7 +1821,7 @@ adw_alert_dialog_get_extra_child (AdwAlertDialog *self)
 }
 
 /**
- * adw_alert_dialog_set_extra_child: (attributes org.gtk.Method.set_property=extra-child)
+ * adw_alert_dialog_set_extra_child:
  * @self: an alert dialog
  * @child: (nullable): the child widget
  *
@@ -1670,17 +1848,69 @@ adw_alert_dialog_set_extra_child (AdwAlertDialog *self,
   if (child == priv->child)
     return;
 
-  if (priv->child)
-    gtk_box_remove (priv->message_area, priv->child);
-
   priv->child = child;
-
-  if (priv->child)
-    gtk_box_append (priv->message_area, priv->child);
+  adw_bin_set_child (ADW_BIN (priv->child_bin), child);
+  gtk_widget_set_visible (priv->child_bin, child != NULL);
 
   g_object_notify_by_pspec (G_OBJECT (self), props[PROP_EXTRA_CHILD]);
 }
 
+/**
+ * adw_alert_dialog_get_prefer_wide_layout:
+ * @self: an alert dialog
+ *
+ * Gets whether @self prefers wide layout.
+ *
+ * Returns: whether to prefer wide layout
+ *
+ * Since: 1.5
+ */
+gboolean
+adw_alert_dialog_get_prefer_wide_layout (AdwAlertDialog *self)
+{
+  AdwAlertDialogPrivate *priv;
+
+  g_return_val_if_fail (ADW_IS_ALERT_DIALOG (self), FALSE);
+
+  priv = adw_alert_dialog_get_instance_private (self);
+
+  return priv->prefer_wide_layout;
+}
+
+/**
+ * adw_alert_dialog_set_prefer_wide_layout:
+ * @self: an alert dialog
+ * @prefer_wide_layout: whether to prefer wide layout
+ *
+ * Sets whether @self prefers wide layout.
+ *
+ * Prefer horizontal button layout when possible, and wider dialog width
+ * otherwise.
+ *
+ * Since: 1.6
+ */
+void
+adw_alert_dialog_set_prefer_wide_layout (AdwAlertDialog *self,
+                                         gboolean        prefer_wide_layout)
+{
+  AdwAlertDialogPrivate *priv;
+
+  g_return_if_fail (ADW_IS_ALERT_DIALOG (self));
+
+  priv = adw_alert_dialog_get_instance_private (self);
+
+  prefer_wide_layout = !!prefer_wide_layout;
+
+  if (prefer_wide_layout == priv->prefer_wide_layout)
+    return;
+
+  priv->prefer_wide_layout = prefer_wide_layout;
+
+  gtk_widget_queue_resize (priv->contents);
+
+  g_object_notify_by_pspec (G_OBJECT (self), props[PROP_PREFER_WIDE_LAYOUT]);
+}
+
 /**
  * adw_alert_dialog_add_response:
  * @self: an alert dialog
@@ -1734,11 +1964,6 @@ adw_alert_dialog_add_response (AdwAlertDialog *self,
   info->appearance = ADW_RESPONSE_DEFAULT;
   info->enabled = TRUE;
 
-  if (priv->responses) {
-    info->separator = gtk_separator_new (GTK_ORIENTATION_VERTICAL);
-    gtk_widget_set_parent (info->separator, priv->response_area);
-  }
-
   info->button = create_response_button (self, info);
   gtk_widget_set_parent (info->button, priv->response_area);
 
@@ -1838,13 +2063,6 @@ adw_alert_dialog_remove_response (AdwAlertDialog *self,
 
   gtk_widget_unparent (info->button);
 
-  if (info == priv->responses->data && priv->responses->next) {
-    ResponseInfo *next_info = priv->responses->next->data;
-    g_clear_pointer (&next_info->separator, gtk_widget_unparent);
-  } else {
-    g_clear_pointer (&info->separator, gtk_widget_unparent);
-  }
-
   priv->responses = g_list_remove (priv->responses, info);
   g_hash_table_remove (priv->id_to_response, id);
 
@@ -1986,14 +2204,14 @@ adw_alert_dialog_set_response_appearance (AdwAlertDialog        *self,
   info->appearance = appearance;
 
   if (info->appearance == ADW_RESPONSE_SUGGESTED)
-    gtk_widget_add_css_class (info->button, "suggested");
+    gtk_widget_add_css_class (info->button, "suggested-action");
   else
-    gtk_widget_remove_css_class (info->button, "suggested");
+    gtk_widget_remove_css_class (info->button, "suggested-action");
 
   if (info->appearance == ADW_RESPONSE_DESTRUCTIVE)
-    gtk_widget_add_css_class (info->button, "destructive");
+    gtk_widget_add_css_class (info->button, "destructive-action");
   else
-    gtk_widget_remove_css_class (info->button, "destructive");
+    gtk_widget_remove_css_class (info->button, "destructive-action");
 }
 
 /**
@@ -2067,7 +2285,7 @@ adw_alert_dialog_set_response_enabled (AdwAlertDialog *self,
 }
 
 /**
- * adw_alert_dialog_get_default_response: (attributes org.gtk.Method.get_property=default-response)
+ * adw_alert_dialog_get_default_response:
  * @self: an alert dialog
  *
  * Gets the ID of the default response of @self.
@@ -2092,7 +2310,7 @@ adw_alert_dialog_get_default_response (AdwAlertDialog *self)
 }
 
 /**
- * adw_alert_dialog_set_default_response: (attributes org.gtk.Method.set_property=default-response)
+ * adw_alert_dialog_set_default_response:
  * @self: an alert dialog
  * @response: (nullable): the default response ID
  *
@@ -2132,7 +2350,7 @@ adw_alert_dialog_set_default_response (AdwAlertDialog *self,
 }
 
 /**
- * adw_alert_dialog_get_close_response: (attributes org.gtk.Method.get_property=close-response)
+ * adw_alert_dialog_get_close_response:
  * @self: an alert dialog
  *
  * Gets the ID of the close response of @self.
@@ -2154,7 +2372,7 @@ adw_alert_dialog_get_close_response (AdwAlertDialog *self)
 }
 
 /**
- * adw_alert_dialog_set_close_response: (attributes org.gtk.Method.set_property=close-response)
+ * adw_alert_dialog_set_close_response:
  * @self: an alert dialog
  * @response: the close response ID
  *
@@ -2236,8 +2454,9 @@ choose_cancelled_cb (GCancellable *cancellable,
                      GTask        *task)
 {
   AdwAlertDialog *self = g_task_get_source_object (task);
+  AdwAlertDialogPrivate *priv = adw_alert_dialog_get_instance_private (self);
 
-  choose_response_cb (self, adw_alert_dialog_get_close_response (self), task);
+  emit_response (self, priv->close_response);
 }
 
 /**
@@ -2250,9 +2469,6 @@ choose_cancelled_cb (GCancellable *cancellable,
  *
  * This function shows @self to the user.
  *
- * The @callback will be called when the alert is dismissed. It should call
- * [method@AlertDialog.choose_finish] to obtain the result.
- *
  * If the window is an [class@Window] or [class@ApplicationWindow], the dialog
  * will be shown within it. Otherwise, it will be a separate window.
  *
@@ -2306,3 +2522,4 @@ adw_alert_dialog_choose_finish (AdwAlertDialog *self,
 
   return g_quark_to_string (id);
 }
+
