@@ -406,22 +406,29 @@ parent_changed_cb (AdwMessageDialog *self)
   set_parent (self, transient_for);
 }
 
-static void
-button_clicked_cb (ResponseInfo *info)
+
+static inline void
+emit_response (AdwMessageDialog *self, GQuark response)
 {
-  AdwMessageDialog *self = info->dialog;
   AdwMessageDialogPrivate *priv = adw_message_dialog_get_instance_private (self);
 
   g_object_ref (self);
   priv->block_close_response = TRUE;
 
   gtk_window_close (GTK_WINDOW (self));
-  g_signal_emit (self, signals[SIGNAL_RESPONSE], info->id, g_quark_to_string (info->id));
+  g_signal_emit (self, signals[SIGNAL_RESPONSE], response, g_quark_to_string (response));
 
   priv->block_close_response = FALSE;
   g_object_unref (self);
 }
 
+
+static void
+button_clicked_cb (ResponseInfo *info)
+{
+  emit_response (info->dialog, info->id);
+}
+
 static GtkWidget *
 create_response_button (AdwMessageDialog *self,
                         ResponseInfo     *info)
@@ -2542,8 +2549,9 @@ choose_cancelled_cb (GCancellable *cancellable,
                      GTask        *task)
 {
   AdwMessageDialog *self = g_task_get_source_object (task);
+  AdwMessageDialogPrivate *priv = adw_message_dialog_get_instance_private (self);
 
-  choose_response_cb (self, adw_message_dialog_get_close_response (self), task);
+  emit_response (self, priv->close_response);
 }
 
 /**
