@@ -66,7 +66,7 @@
  * at once, potentially across multiple navigation views.
  *
  * Set [property@HeaderBar:show-back-button] to `FALSE` to disable this behavior
- * if it's unwanted.
+ * in rare scenarios where it's unwanted.
  *
  * `AdwHeaderBar` will also display the title of the `AdwNavigationPage` it's
  * placed into, so most applications shouldn't need to customize it at all.
@@ -227,6 +227,8 @@ typedef struct
   int block_signals;
 
   AdwNavigationView *child_view;
+
+  int nav_split_views;
 } AdwNavigationPagePrivate;
 
 static void adw_navigation_page_buildable_init (GtkBuildableIface *iface);
@@ -275,6 +277,7 @@ struct _AdwNavigationView
   gboolean transition_cancel;
   double transition_progress;
   gboolean gesture_active;
+  AdwNavigationDirection swipe_direction;
 
   AdwShadowHelper *shadow_helper;
   AdwSwipeTracker *swipe_tracker;
@@ -403,7 +406,7 @@ adw_navigation_page_realize (GtkWidget *widget)
 
   GTK_WIDGET_CLASS (adw_navigation_page_parent_class)->realize (widget);
 
-  if (!(priv->title && *priv->title) && !priv->child_view) {
+  if (!(priv->title && *priv->title) && !priv->child_view && priv->nav_split_views == 0) {
     g_warning ("AdwNavigationPage %p is missing a title. To hide a header bar " \
                "title, consider using AdwHeaderBar:show-title instead.", self);
   }
@@ -416,13 +419,7 @@ adw_navigation_page_dispose (GObject *object)
   AdwNavigationPagePrivate *priv = adw_navigation_page_get_instance_private (self);
 
   g_clear_pointer (&priv->child, gtk_widget_unparent);
-
-  if (priv->child_view) {
-    g_object_remove_weak_pointer (G_OBJECT (priv->child_view),
-                                  (gpointer *) &priv->child_view);
-
-    priv->child_view = NULL;
-  }
+  g_clear_weak_pointer (&priv->child_view);
 
   G_OBJECT_CLASS (adw_navigation_page_parent_class)->dispose (object);
 }
@@ -435,10 +432,7 @@ adw_navigation_page_finalize (GObject *object)
 
   g_free (priv->title);
   g_free (priv->tag);
-
-  if (priv->last_focus)
-    g_object_remove_weak_pointer (G_OBJECT (priv->last_focus),
-                                  (gpointer *) &priv->last_focus);
+  g_clear_weak_pointer (&priv->last_focus);
 
   G_OBJECT_CLASS (adw_navigation_page_parent_class)->finalize (object);
 }
@@ -510,7 +504,7 @@ adw_navigation_page_class_init (AdwNavigationPageClass *klass)
   widget_class->compute_expand = adw_widget_compute_expand;
 
   /**
-   * AdwNavigationPage:child: (attributes org.gtk.Property.get=adw_navigation_page_get_child org.gtk.Property.set=adw_navigation_page_set_child)
+   * AdwNavigationPage:child:
    *
    * The child widget.
    *
@@ -522,7 +516,7 @@ adw_navigation_page_class_init (AdwNavigationPageClass *klass)
                          G_PARAM_READWRITE | G_PARAM_STATIC_STRINGS | G_PARAM_EXPLICIT_NOTIFY);
 
   /**
-   * AdwNavigationPage:tag: (attributes org.gtk.Property.get=adw_navigation_page_get_tag org.gtk.Property.set=adw_navigation_page_set_tag)
+   * AdwNavigationPage:tag:
    *
    * The page tag.
    *
@@ -543,7 +537,7 @@ adw_navigation_page_class_init (AdwNavigationPageClass *klass)
                          G_PARAM_READWRITE | G_PARAM_STATIC_STRINGS | G_PARAM_EXPLICIT_NOTIFY);
 
   /**
-   * AdwNavigationPage:title: (attributes org.gtk.Property.get=adw_navigation_page_get_title org.gtk.Property.set=adw_navigation_page_set_title)
+   * AdwNavigationPage:title:
    *
    * The page title.
    *
@@ -558,7 +552,7 @@ adw_navigation_page_class_init (AdwNavigationPageClass *klass)
                          G_PARAM_READWRITE | G_PARAM_STATIC_STRINGS | G_PARAM_EXPLICIT_NOTIFY);
 
   /**
-   * AdwNavigationPage:can-pop: (attributes org.gtk.Property.get=adw_navigation_page_get_can_pop org.gtk.Property.set=adw_navigation_page_set_can_pop)
+   * AdwNavigationPage:can-pop:
    *
    * Whether the page can be popped from navigation stack.
    *
@@ -744,12 +738,7 @@ switch_page (AdwNavigationView *self,
 
     contains_focus = TRUE;
 
-    if (priv->last_focus)
-      g_object_remove_weak_pointer (G_OBJECT (priv->last_focus),
-                                    (gpointer *)&priv->last_focus);
-    priv->last_focus = focus;
-    g_object_add_weak_pointer (G_OBJECT (priv->last_focus),
-                               (gpointer *)&priv->last_focus);
+    g_set_weak_pointer (&priv->last_focus, focus);
   }
 
   if (!prev_page)
@@ -1221,22 +1210,41 @@ static void
 prepare_cb (AdwSwipeTracker        *tracker,
             AdwNavigationDirection  direction,
             AdwNavigationView      *self)
+{
+  if (!adw_navigation_view_get_visible_page (self))
+    return;
+
+  self->swipe_direction = ADW_NAVIGATION_DIRECTION_BACK;
+}
+
+static void
+begin_swipe_cb (AdwSwipeTracker   *tracker,
+                AdwNavigationView *self)
 {
   AdwNavigationPage *visible_page = adw_navigation_view_get_visible_page (self);
   AdwNavigationPage *new_page;
   gboolean remove_on_pop = FALSE;
 
-  if (!visible_page)
+  if (self->swipe_direction < 0)
     return;
 
-  if (direction == ADW_NAVIGATION_DIRECTION_BACK) {
-    if (!adw_navigation_page_get_can_pop (visible_page))
+  if (!visible_page) {
+    self->swipe_direction = -1;
+    return;
+  }
+
+  if (self->swipe_direction == ADW_NAVIGATION_DIRECTION_BACK) {
+    if (!adw_navigation_page_get_can_pop (visible_page)) {
+      self->swipe_direction = -1;
       return;
+    }
 
     new_page = adw_navigation_view_get_previous_page (self, visible_page);
 
-    if (!new_page)
+    if (!new_page) {
+      self->swipe_direction = -1;
       return;
+    }
 
   } else {
     new_page = get_next_page (self);
@@ -1254,9 +1262,9 @@ prepare_cb (AdwSwipeTracker        *tracker,
   self->showing_page = new_page;
   self->hiding_page = g_object_ref (visible_page);
 
-  self->transition_pop = (direction == ADW_NAVIGATION_DIRECTION_BACK);
+  self->transition_pop = (self->swipe_direction == ADW_NAVIGATION_DIRECTION_BACK);
 
-  if (direction == ADW_NAVIGATION_DIRECTION_BACK) {
+  if (self->swipe_direction == ADW_NAVIGATION_DIRECTION_BACK) {
     g_object_ref (new_page);
   } else {
     if (remove_on_pop)
@@ -1281,6 +1289,8 @@ prepare_cb (AdwSwipeTracker        *tracker,
   gtk_widget_queue_resize (GTK_WIDGET (self));
 
   adw_swipe_tracker_set_upper_overshoot (self->swipe_tracker, TRUE);
+
+  self->swipe_direction = -1;
 }
 
 static void
@@ -1307,6 +1317,8 @@ end_swipe_cb (AdwSwipeTracker   *tracker,
 {
   gboolean animate;
 
+  self->swipe_direction = -1;
+
   if (!self->gesture_active)
     return;
 
@@ -1355,20 +1367,7 @@ set_child_view (AdwNavigationPage *self,
 {
   AdwNavigationPagePrivate *priv = adw_navigation_page_get_instance_private (self);
 
-  if (view == priv->child_view)
-    return;
-
-  if (priv->child_view) {
-    g_object_remove_weak_pointer (G_OBJECT (priv->child_view),
-                                  (gpointer *) &priv->child_view);
-  }
-
-  priv->child_view = view;
-
-  if (priv->child_view) {
-    g_object_add_weak_pointer (G_OBJECT (priv->child_view),
-                               (gpointer *) &priv->child_view);
-  }
+  g_set_weak_pointer (&priv->child_view, view);
 }
 
 static void
@@ -1571,7 +1570,7 @@ adw_navigation_view_root (GtkWidget *widget)
 
   GTK_WIDGET_CLASS (adw_navigation_view_parent_class)->root (widget);
 
-  parent_page = adw_widget_get_ancestor_same_native (widget, ADW_TYPE_NAVIGATION_PAGE);
+  parent_page = adw_widget_get_ancestor (widget, ADW_TYPE_NAVIGATION_PAGE, TRUE, TRUE);
 
   if (parent_page)
     set_child_view (ADW_NAVIGATION_PAGE (parent_page), self);
@@ -1582,7 +1581,7 @@ adw_navigation_view_unroot (GtkWidget *widget)
 {
   GtkWidget *parent_page;
 
-  parent_page = adw_widget_get_ancestor_same_native (widget, ADW_TYPE_NAVIGATION_PAGE);
+  parent_page = adw_widget_get_ancestor (widget, ADW_TYPE_NAVIGATION_PAGE, TRUE, TRUE);
 
   if (parent_page)
     set_child_view (ADW_NAVIGATION_PAGE (parent_page), NULL);
@@ -1630,9 +1629,7 @@ adw_navigation_view_finalize (GObject *object)
 {
   AdwNavigationView *self = ADW_NAVIGATION_VIEW (object);
 
-  if (self->navigation_stack_model)
-    g_object_remove_weak_pointer (G_OBJECT (self->navigation_stack_model),
-                                  (gpointer *) &self->navigation_stack_model);
+  g_clear_weak_pointer (&self->navigation_stack_model);
 
   G_OBJECT_CLASS (adw_navigation_view_parent_class)->finalize (object);
 }
@@ -1717,7 +1714,7 @@ adw_navigation_view_class_init (AdwNavigationViewClass *klass)
   widget_class->compute_expand = adw_widget_compute_expand;
 
   /**
-   * AdwNavigationView:visible-page: (attributes org.gtk.Property.get=adw_navigation_view_get_visible_page)
+   * AdwNavigationView:visible-page:
    *
    * The currently visible page.
    *
@@ -1729,7 +1726,7 @@ adw_navigation_view_class_init (AdwNavigationViewClass *klass)
                          G_PARAM_READABLE | G_PARAM_STATIC_STRINGS);
 
   /**
-   * AdwNavigationView:animate-transitions: (attributes org.gtk.Property.get=adw_navigation_view_get_animate_transitions org.gtk.Property.set=adw_navigation_view_set_animate_transitions)
+   * AdwNavigationView:animate-transitions:
    *
    * Whether to animate page transitions.
    *
@@ -1743,7 +1740,7 @@ adw_navigation_view_class_init (AdwNavigationViewClass *klass)
                           G_PARAM_READWRITE | G_PARAM_STATIC_STRINGS | G_PARAM_EXPLICIT_NOTIFY);
 
   /**
-   * AdwNavigationView:pop-on-escape: (attributes org.gtk.Property.get=adw_navigation_view_get_pop_on_escape org.gtk.Property.set=adw_navigation_view_set_pop_on_escape)
+   * AdwNavigationView:pop-on-escape:
    *
    * Whether pressing Escape pops the current page.
    *
@@ -1758,7 +1755,7 @@ adw_navigation_view_class_init (AdwNavigationViewClass *klass)
                           G_PARAM_READWRITE | G_PARAM_STATIC_STRINGS | G_PARAM_EXPLICIT_NOTIFY);
 
   /**
-   * AdwNavigationView:navigation-stack: (attributes org.gtk.Property.get=adw_navigation_view_get_navigation_stack)
+   * AdwNavigationView:navigation-stack:
    *
    * A list model that contains the pages in navigation stack.
    *
@@ -1938,11 +1935,15 @@ adw_navigation_view_init (AdwNavigationView *self)
 
   g_signal_connect (self->swipe_tracker, "prepare",
                     G_CALLBACK (prepare_cb), self);
+  g_signal_connect (self->swipe_tracker, "begin-swipe",
+                    G_CALLBACK (begin_swipe_cb), self);
   g_signal_connect (self->swipe_tracker, "update-swipe",
                     G_CALLBACK (update_swipe_cb), self);
   g_signal_connect (self->swipe_tracker, "end-swipe",
                     G_CALLBACK (end_swipe_cb), self);
 
+  self->swipe_direction = -1;
+
   self->shield = adw_gizmo_new ("widget", NULL, NULL, NULL, NULL, NULL, NULL);
   gtk_widget_set_child_visible (self->shield, FALSE);
   gtk_widget_set_parent (self->shield, GTK_WIDGET (self));
@@ -2095,7 +2096,7 @@ adw_navigation_page_new_with_tag (GtkWidget  *child,
 }
 
 /**
- * adw_navigation_page_get_child: (attributes org.gtk.Method.get_property=child)
+ * adw_navigation_page_get_child:
  * @self: a navigation page
  *
  * Gets the child widget of @self.
@@ -2117,7 +2118,7 @@ adw_navigation_page_get_child (AdwNavigationPage *self)
 }
 
 /**
- * adw_navigation_page_set_child: (attributes org.gtk.Method.set_property=child)
+ * adw_navigation_page_set_child:
  * @self: a navigation page
  * @child: (nullable): the child widget
  *
@@ -2158,7 +2159,7 @@ adw_navigation_page_set_child (AdwNavigationPage *self,
 }
 
 /**
- * adw_navigation_page_get_tag: (attributes org.gtk.Method.get_property=tag)
+ * adw_navigation_page_get_tag:
  * @self: a navigation page
  *
  * Gets the tag of @self.
@@ -2180,7 +2181,7 @@ adw_navigation_page_get_tag (AdwNavigationPage *self)
 }
 
 /**
- * adw_navigation_page_set_tag: (attributes org.gtk.Method.set_property=tag)
+ * adw_navigation_page_set_tag:
  * @self: a navigation page
  * @tag: (nullable): the page tag
  *
@@ -2236,7 +2237,7 @@ adw_navigation_page_set_tag (AdwNavigationPage *self,
 }
 
 /**
- * adw_navigation_page_get_title: (attributes org.gtk.Method.get_property=title)
+ * adw_navigation_page_get_title:
  * @self: a navigation page
  *
  * Gets the title of @self.
@@ -2258,7 +2259,7 @@ adw_navigation_page_get_title (AdwNavigationPage *self)
 }
 
 /**
- * adw_navigation_page_set_title: (attributes org.gtk.Method.set_property=title)
+ * adw_navigation_page_set_title:
  * @self: a navigation page
  * @title: the title
  *
@@ -2291,7 +2292,7 @@ adw_navigation_page_set_title (AdwNavigationPage *self,
 }
 
 /**
- * adw_navigation_page_get_can_pop: (attributes org.gtk.Method.get_property=can-pop)
+ * adw_navigation_page_get_can_pop:
  * @self: a navigation page
  *
  * Gets whether @self can be popped from navigation stack.
@@ -2313,7 +2314,7 @@ adw_navigation_page_get_can_pop (AdwNavigationPage *self)
 }
 
 /**
- * adw_navigation_page_set_can_pop: (attributes org.gtk.Method.set_property=can-pop)
+ * adw_navigation_page_set_can_pop:
  * @self: a navigation page
  * @can_pop: whether the page can be popped from navigation stack
  *
@@ -2428,6 +2429,26 @@ adw_navigation_page_unblock_signals (AdwNavigationPage *self)
   priv->block_signals--;
 }
 
+void
+adw_navigation_page_add_child_nav_split_view (AdwNavigationPage *self)
+{
+  AdwNavigationPagePrivate *priv = adw_navigation_page_get_instance_private (self);
+
+  g_return_if_fail (ADW_IS_NAVIGATION_PAGE (self));
+
+  priv->nav_split_views++;
+}
+
+void
+adw_navigation_page_remove_child_nav_split_view (AdwNavigationPage *self)
+{
+  AdwNavigationPagePrivate *priv = adw_navigation_page_get_instance_private (self);
+
+  g_return_if_fail (ADW_IS_NAVIGATION_PAGE (self));
+
+  priv->nav_split_views--;
+}
+
 /**
  * adw_navigation_view_new:
  *
@@ -2891,7 +2912,7 @@ adw_navigation_view_replace_with_tags (AdwNavigationView  *self,
 }
 
 /**
- * adw_navigation_view_get_visible_page: (attributes org.gtk.Method.get_property=visible-page)
+ * adw_navigation_view_get_visible_page:
  * @self: a navigation view
  *
  * Gets the currently visible page in @self.
@@ -2960,7 +2981,7 @@ adw_navigation_view_get_previous_page (AdwNavigationView *self,
 }
 
 /**
- * adw_navigation_view_get_animate_transitions: (attributes org.gtk.Method.get_property=animate-transitions)
+ * adw_navigation_view_get_animate_transitions:
  * @self: a navigation view
  *
  * Gets whether @self animates page transitions.
@@ -2978,7 +2999,7 @@ adw_navigation_view_get_animate_transitions (AdwNavigationView *self)
 }
 
 /**
- * adw_navigation_view_set_animate_transitions: (attributes org.gtk.Method.set_property=animate-transitions)
+ * adw_navigation_view_set_animate_transitions:
  * @self: a navigation view
  * @animate_transitions: whether to animate page transitions
  *
@@ -3005,7 +3026,7 @@ adw_navigation_view_set_animate_transitions (AdwNavigationView *self,
 }
 
 /**
- * adw_navigation_view_get_pop_on_escape: (attributes org.gtk.Method.get_property=pop-on-escape)
+ * adw_navigation_view_get_pop_on_escape:
  * @self: a navigation view
  *
  * Gets whether pressing Escape pops the current page on @self.
@@ -3023,7 +3044,7 @@ adw_navigation_view_get_pop_on_escape (AdwNavigationView *self)
 }
 
 /**
- * adw_navigation_view_set_pop_on_escape: (attributes org.gtk.Method.set_property=pop-on-escape)
+ * adw_navigation_view_set_pop_on_escape:
  * @self: a navigation view
  * @pop_on_escape: whether to pop the current page when pressing Escape
  *
@@ -3051,7 +3072,7 @@ adw_navigation_view_set_pop_on_escape (AdwNavigationView *self,
 }
 
 /**
- * adw_navigation_view_get_navigation_stack: (attributes org.gtk.Method.get_property=navigation-stack)
+ * adw_navigation_view_get_navigation_stack:
  * @self: a navigation view
  *
  * Returns a [iface@Gio.ListModel] that contains the pages in navigation stack.
@@ -3072,9 +3093,8 @@ adw_navigation_view_get_navigation_stack (AdwNavigationView *self)
   if (self->navigation_stack_model)
     return g_object_ref (self->navigation_stack_model);
 
-  self->navigation_stack_model = adw_navigation_view_model_new (self);
-  g_object_add_weak_pointer (G_OBJECT (self->navigation_stack_model),
-                             (gpointer *) &self->navigation_stack_model);
+  g_set_weak_pointer (&self->navigation_stack_model,
+                      adw_navigation_view_model_new (self));
 
   return self->navigation_stack_model;
 }
