@@ -69,9 +69,10 @@ typedef struct
 
   gboolean search_enabled;
 
-  GtkFilter *row_filter;
-  GtkFilter *page_filter;
+  GtkFilter *filter;
   GtkFilterListModel *filter_model;
+
+  int n_pages;
 } AdwPreferencesDialogPrivate;
 
 static void adw_preferences_dialog_buildable_init (GtkBuildableIface *iface);
@@ -93,6 +94,50 @@ enum {
 
 static GParamSpec *props[LAST_PROP];
 
+/* Copied and modified from gtklabel.c, separate_uline_pattern() */
+static char *
+strip_mnemonic (const char *src)
+{
+  char *new_str = g_new (char, strlen (src) + 1);
+  char *dest = new_str;
+  gboolean underscore = FALSE;
+
+  while (*src) {
+    gunichar c;
+    const char *next_src;
+
+    c = g_utf8_get_char (src);
+    if (c == (gunichar) -1) {
+      g_warning ("Invalid input string");
+
+      g_free (new_str);
+
+      return NULL;
+    }
+
+    next_src = g_utf8_next_char (src);
+
+    if (underscore) {
+      while (src < next_src)
+        *dest++ = *src++;
+
+      underscore = FALSE;
+    } else {
+      if (c == '_'){
+        underscore = TRUE;
+        src = next_src;
+      } else {
+        while (src < next_src)
+          *dest++ = *src++;
+      }
+    }
+  }
+
+  *dest = 0;
+
+  return new_str;
+}
+
 static char *
 make_comparable (const char        *src,
                  AdwPreferencesRow *row,
@@ -114,7 +159,7 @@ make_comparable (const char        *src,
   }
 
   if (allow_underline && adw_preferences_row_get_use_underline (row)) {
-    char *comparable = adw_strip_mnemonic (plaintext);
+    char *comparable = strip_mnemonic (plaintext);
     g_free (plaintext);
     return comparable;
   }
@@ -194,7 +239,7 @@ create_search_row_subtitle (AdwPreferencesDialog *self,
     const char *title = adw_preferences_page_get_title (ADW_PREFERENCES_PAGE (page));
 
     if (adw_preferences_page_get_use_underline (ADW_PREFERENCES_PAGE (page)))
-      page_title = adw_strip_mnemonic (title);
+      page_title = strip_mnemonic (title);
     else
       page_title = g_strdup (title);
 
@@ -309,11 +354,10 @@ update_view_switcher (AdwPreferencesDialog *self)
   AdwPreferencesDialogPrivate *priv = adw_preferences_dialog_get_instance_private (self);
   AdwBreakpoint *breakpoint;
   AdwBreakpointCondition *main_condition, *fallback_condition, *condition;
-  int n_pages = get_n_pages (self);
 
   main_condition =
     adw_breakpoint_condition_new_length (ADW_BREAKPOINT_CONDITION_MAX_WIDTH,
-                                         VIEW_SWITCHER_PAGE_THRESHOLD * MAX (1, n_pages),
+                                         VIEW_SWITCHER_PAGE_THRESHOLD * MAX (1, priv->n_pages),
                                          ADW_LENGTH_UNIT_PT);
   fallback_condition =
     adw_breakpoint_condition_new_length (ADW_BREAKPOINT_CONDITION_MAX_WIDTH,
@@ -326,14 +370,12 @@ update_view_switcher (AdwPreferencesDialog *self)
 
   breakpoint = adw_breakpoint_bin_get_current_breakpoint (ADW_BREAKPOINT_BIN (priv->breakpoint_bin));
 
-  if (!breakpoint && n_pages > 1)
+  if (!breakpoint && priv->n_pages > 1)
     gtk_stack_set_visible_child (GTK_STACK (priv->view_switcher_stack), priv->view_switcher);
   else
     gtk_stack_set_visible_child (GTK_STACK (priv->view_switcher_stack), priv->title);
 
   adw_breakpoint_condition_free (condition);
-
-  gtk_filter_changed (priv->page_filter, GTK_FILTER_CHANGE_DIFFERENT);
 }
 
 static void
@@ -405,7 +447,7 @@ search_changed_cb (AdwPreferencesDialog *self)
   AdwPreferencesDialogPrivate *priv = adw_preferences_dialog_get_instance_private (self);
   guint n;
 
-  gtk_filter_changed (priv->row_filter, GTK_FILTER_CHANGE_DIFFERENT);
+  gtk_filter_changed (priv->filter, GTK_FILTER_CHANGE_DIFFERENT);
 
   n = g_list_model_get_n_items (G_LIST_MODEL (priv->filter_model));
 
@@ -577,7 +619,6 @@ adw_preferences_dialog_class_init (AdwPreferencesDialogClass *klass)
   gtk_widget_class_bind_template_callback (widget_class, search_results_map);
   gtk_widget_class_bind_template_callback (widget_class, search_results_unmap);
   gtk_widget_class_bind_template_callback (widget_class, stop_search_cb);
-  gtk_widget_class_bind_template_callback (widget_class, adw_tab_on_vertical_keynav_failed);
 }
 
 static GListModel *
@@ -601,19 +642,17 @@ adw_preferences_dialog_init (AdwPreferencesDialog *self)
 
   gtk_widget_init_template (GTK_WIDGET (self));
 
-  priv->row_filter = GTK_FILTER (gtk_custom_filter_new ((GtkCustomFilterFunc) filter_search_results, self, NULL));
-
+  priv->filter = GTK_FILTER (gtk_custom_filter_new ((GtkCustomFilterFunc) filter_search_results, self, NULL));
   expr = gtk_property_expression_new (ADW_TYPE_VIEW_STACK_PAGE, NULL, "visible");
-  priv->page_filter = GTK_FILTER (gtk_bool_filter_new (expr));
 
   model = G_LIST_MODEL (adw_view_stack_get_pages (priv->pages_stack));
-  model = G_LIST_MODEL (gtk_filter_list_model_new (model, priv->page_filter));
+  model = G_LIST_MODEL (gtk_filter_list_model_new (model, GTK_FILTER (gtk_bool_filter_new (expr))));
   model = G_LIST_MODEL (gtk_map_list_model_new (model,
                                                 (GtkMapListModelMapFunc) preferences_page_to_rows,
                                                 NULL,
                                                 NULL));
   model = G_LIST_MODEL (gtk_flatten_list_model_new (model));
-  priv->filter_model = gtk_filter_list_model_new (model, priv->row_filter);
+  priv->filter_model = gtk_filter_list_model_new (model, priv->filter);
 
   gtk_widget_set_visible (GTK_WIDGET (priv->search_button), FALSE);
 }
@@ -684,10 +723,8 @@ adw_preferences_dialog_add (AdwPreferencesDialog *self,
   g_object_bind_property (page, "title", stack_page, "title", G_BINDING_SYNC_CREATE);
   g_object_bind_property (page, "use-underline", stack_page, "use-underline", G_BINDING_SYNC_CREATE);
   g_object_bind_property (page, "name", stack_page, "name", G_BINDING_SYNC_CREATE);
-  g_object_bind_property (page, "visible", stack_page, "visible", G_BINDING_SYNC_CREATE);
-
-  g_signal_connect_swapped (stack_page, "notify::visible", G_CALLBACK (update_view_switcher), self);
 
+  priv->n_pages++;
   update_view_switcher (self);
 }
 
@@ -716,6 +753,7 @@ adw_preferences_dialog_remove (AdwPreferencesDialog *self,
   else
     ADW_CRITICAL_CANNOT_REMOVE_CHILD (self, page);
 
+  priv->n_pages--;
   update_view_switcher (self);
 }
 
