@@ -11,15 +11,18 @@
 
 #include "adw-animation-util.h"
 #include "adw-action-row.h"
+#include "adw-breakpoint-bin.h"
 #include "adw-leaflet.h"
+#include "adw-navigation-view.h"
 #include "adw-preferences-group-private.h"
 #include "adw-preferences-page-private.h"
 #include "adw-toast-overlay.h"
-#include "adw-view-switcher.h"
-#include "adw-view-switcher-bar.h"
-#include "adw-view-switcher-title.h"
+#include "adw-view-stack.h"
 #include "adw-widget-utils-private.h"
 
+#define VIEW_SWITCHER_PAGE_THRESHOLD 110
+#define VIEW_SWITCHER_FALLBACK_THRESHOLD 400
+
 /**
  * AdwPreferencesWindow:
  *
@@ -37,13 +40,15 @@
  *
  * `AdwPreferencesWindow` has a main CSS node with the name `window` and the
  * style class `.preferences`.
+ *
+ * Deprecated: 1.6: Use [class@PreferencesDialog].
  */
 
 typedef struct
 {
   AdwToastOverlay *toast_overlay;
-  AdwLeaflet *subpages_leaflet;
-  GtkWidget *preferences;
+  AdwNavigationView *subpages_nav_view;
+  GtkWidget *breakpoint_bin;
   GtkStack *content_stack;
   AdwViewStack *pages_stack;
   GtkToggleButton *search_button;
@@ -51,8 +56,10 @@ typedef struct
   GtkListBox *search_results;
   GtkStack *search_stack;
   GtkStack *title_stack;
-  AdwViewSwitcherBar *view_switcher_bar;
-  AdwViewSwitcherTitle *view_switcher_title;
+  GtkWidget *view_switcher_stack;
+  GtkWidget *view_switcher;
+  GtkWidget *title;
+  AdwBreakpoint *breakpoint;
 
   gboolean search_enabled;
   gboolean can_navigate_back;
@@ -60,6 +67,9 @@ typedef struct
   GtkFilter *filter;
   GtkFilterListModel *filter_model;
 
+  int n_pages;
+
+  AdwLeaflet *subpages_leaflet;
   GtkWidget *subpage;
 } AdwPreferencesWindowPrivate;
 
@@ -333,7 +343,7 @@ search_results_unmap (AdwPreferencesWindow *self)
 }
 
 static void
-try_remove_subpages (AdwPreferencesWindow *self)
+try_remove_legacy_subpages (AdwPreferencesWindow *self)
 {
   AdwPreferencesWindowPrivate *priv = adw_preferences_window_get_instance_private (self);
   GtkWidget *child;
@@ -341,7 +351,7 @@ try_remove_subpages (AdwPreferencesWindow *self)
   if (adw_leaflet_get_child_transition_running (priv->subpages_leaflet))
     return;
 
-  if (adw_leaflet_get_visible_child (priv->subpages_leaflet) == priv->preferences)
+  if (adw_leaflet_get_visible_child (priv->subpages_leaflet) == GTK_WIDGET (priv->subpages_nav_view))
     priv->subpage = NULL;
 
   child = gtk_widget_get_first_child (GTK_WIDGET (priv->subpages_leaflet));
@@ -350,21 +360,39 @@ try_remove_subpages (AdwPreferencesWindow *self)
 
     child = gtk_widget_get_next_sibling (child);
 
-    if (page != priv->preferences && page != priv->subpage)
+    if (page != GTK_WIDGET (priv->subpages_nav_view) && page != priv->subpage)
       adw_leaflet_remove (priv->subpages_leaflet, page);
   }
 }
 
 static void
-subpages_leaflet_child_transition_running_cb (AdwPreferencesWindow *self)
+update_view_switcher (AdwPreferencesWindow *self)
 {
-  try_remove_subpages (self);
-}
+  AdwPreferencesWindowPrivate *priv = adw_preferences_window_get_instance_private (self);
+  AdwBreakpoint *breakpoint;
+  AdwBreakpointCondition *main_condition, *fallback_condition, *condition;
 
-static void
-subpages_leaflet_visible_child_cb (AdwPreferencesWindow *self)
-{
-  try_remove_subpages (self);
+  main_condition =
+    adw_breakpoint_condition_new_length (ADW_BREAKPOINT_CONDITION_MAX_WIDTH,
+                                         VIEW_SWITCHER_PAGE_THRESHOLD * MAX (1, priv->n_pages),
+                                         ADW_LENGTH_UNIT_PT);
+  fallback_condition =
+    adw_breakpoint_condition_new_length (ADW_BREAKPOINT_CONDITION_MAX_WIDTH,
+                                         VIEW_SWITCHER_FALLBACK_THRESHOLD,
+                                         ADW_LENGTH_UNIT_PX);
+
+  condition = adw_breakpoint_condition_new_or (main_condition, fallback_condition);
+
+  adw_breakpoint_set_condition (priv->breakpoint, condition);
+
+  breakpoint = adw_breakpoint_bin_get_current_breakpoint (ADW_BREAKPOINT_BIN (priv->breakpoint_bin));
+
+  if (!breakpoint && priv->n_pages > 1)
+    gtk_stack_set_visible_child (GTK_STACK (priv->view_switcher_stack), priv->view_switcher);
+  else
+    gtk_stack_set_visible_child (GTK_STACK (priv->view_switcher_stack), priv->title);
+
+  adw_breakpoint_condition_free (condition);
 }
 
 static void
@@ -373,7 +401,7 @@ title_stack_notify_transition_running_cb (AdwPreferencesWindow *self)
   AdwPreferencesWindowPrivate *priv = adw_preferences_window_get_instance_private (self);
 
   if (gtk_stack_get_transition_running (priv->title_stack) ||
-      gtk_stack_get_visible_child (priv->title_stack) != GTK_WIDGET (priv->view_switcher_title))
+      gtk_stack_get_visible_child (priv->title_stack) != priv->view_switcher_stack)
     return;
 
   gtk_editable_set_text (GTK_EDITABLE (priv->search_entry), "");
@@ -385,7 +413,7 @@ title_stack_notify_visible_child_cb (AdwPreferencesWindow *self)
   AdwPreferencesWindowPrivate *priv = adw_preferences_window_get_instance_private (self);
 
   if (adw_get_enable_animations (GTK_WIDGET (priv->title_stack)) ||
-      gtk_stack_get_visible_child (priv->title_stack) != GTK_WIDGET (priv->view_switcher_title))
+      gtk_stack_get_visible_child (priv->title_stack) != priv->view_switcher_stack)
     return;
 
   gtk_editable_set_text (GTK_EDITABLE (priv->search_entry), "");
@@ -563,9 +591,11 @@ adw_preferences_window_class_init (AdwPreferencesWindowClass *klass)
   object_class->dispose = adw_preferences_window_dispose;
 
   /**
-   * AdwViewStack:visible-page: (attributes org.gtk.Property.get=adw_preferences_window_get_visible_page org.gtk.Property.set=adw_preferences_window_set_visible_page)
+   * AdwPreferencesWindow:visible-page:
    *
    * The currently visible page.
+   *
+   * Deprecated: 1.6: Use [class@PreferencesDialog].
    */
   props[PROP_VISIBLE_PAGE] =
     g_param_spec_object ("visible-page", NULL, NULL,
@@ -573,11 +603,13 @@ adw_preferences_window_class_init (AdwPreferencesWindowClass *klass)
                          G_PARAM_READWRITE | G_PARAM_STATIC_STRINGS | G_PARAM_EXPLICIT_NOTIFY);
 
   /**
-   * AdwViewStack:visible-page-name: (attributes org.gtk.Property.get=adw_preferences_window_get_visible_page_name org.gtk.Property.set=adw_preferences_window_set_visible_page_name)
+   * AdwPreferencesWindow:visible-page-name:
    *
    * The name of the currently visible page.
    *
-   * See [property@ViewStack:visible-child].
+   * See [property@PreferencesWindow:visible-page].
+   *
+   * Deprecated: 1.6: Use [class@PreferencesDialog].
    */
   props[PROP_VISIBLE_PAGE_NAME] =
     g_param_spec_string ("visible-page-name", NULL, NULL,
@@ -585,9 +617,11 @@ adw_preferences_window_class_init (AdwPreferencesWindowClass *klass)
                          G_PARAM_READWRITE | G_PARAM_STATIC_STRINGS | G_PARAM_EXPLICIT_NOTIFY);
 
   /**
-   * AdwPreferencesWindow:search-enabled: (attributes org.gtk.Property.get=adw_preferences_window_get_search_enabled org.gtk.Property.set=adw_preferences_window_set_search_enabled)
+   * AdwPreferencesWindow:search-enabled:
    *
    * Whether search is enabled.
+   *
+   * Deprecated: 1.6: Use [class@PreferencesDialog].
    */
   props[PROP_SEARCH_ENABLED] =
     g_param_spec_boolean ("search-enabled", NULL, NULL,
@@ -595,7 +629,7 @@ adw_preferences_window_class_init (AdwPreferencesWindowClass *klass)
                           G_PARAM_READWRITE | G_PARAM_STATIC_STRINGS | G_PARAM_EXPLICIT_NOTIFY);
 
   /**
-   * AdwPreferencesWindow:can-navigate-back: (attributes org.gtk.Property.get=adw_preferences_window_get_can_navigate_back org.gtk.Property.set=adw_preferences_window_set_can_navigate_back)
+   * AdwPreferencesWindow:can-navigate-back:
    *
    * Whether gestures and shortcuts for closing subpages are enabled.
    *
@@ -609,22 +643,33 @@ adw_preferences_window_class_init (AdwPreferencesWindowClass *klass)
    * <kbd>Alt</kbd>+<kbd>←</kbd> shortcut.
    *
    * For right-to-left locales, gestures and shortcuts are reversed.
+   *
+   * Has no effect for subpages added with
+   * [method@PreferencesWindow.push_subpage].
+   *
+   * Deprecated: 1.4: Use [property@NavigationPage:can-pop] instead.
    */
   props[PROP_CAN_NAVIGATE_BACK] =
     g_param_spec_boolean ("can-navigate-back", NULL, NULL,
                           FALSE,
-                          G_PARAM_READWRITE | G_PARAM_STATIC_STRINGS | G_PARAM_EXPLICIT_NOTIFY);
+                          G_PARAM_READWRITE | G_PARAM_STATIC_STRINGS | G_PARAM_EXPLICIT_NOTIFY | G_PARAM_DEPRECATED);
 
   g_object_class_install_properties (object_class, LAST_PROP, props);
 
+#ifdef __APPLE__
+  gtk_widget_class_add_binding (widget_class, GDK_KEY_f, GDK_META_MASK, search_open_cb, NULL);
+#else
   gtk_widget_class_add_binding (widget_class, GDK_KEY_f, GDK_CONTROL_MASK, search_open_cb, NULL);
+#endif
+
   gtk_widget_class_add_binding (widget_class, GDK_KEY_Escape, 0, close_cb, NULL);
 
   gtk_widget_class_set_template_from_resource (widget_class,
                                                "/org/gnome/Adwaita/ui/adw-preferences-window.ui");
   gtk_widget_class_bind_template_child_private (widget_class, AdwPreferencesWindow, toast_overlay);
   gtk_widget_class_bind_template_child_private (widget_class, AdwPreferencesWindow, subpages_leaflet);
-  gtk_widget_class_bind_template_child_private (widget_class, AdwPreferencesWindow, preferences);
+  gtk_widget_class_bind_template_child_private (widget_class, AdwPreferencesWindow, subpages_nav_view);
+  gtk_widget_class_bind_template_child_private (widget_class, AdwPreferencesWindow, breakpoint_bin);
   gtk_widget_class_bind_template_child_private (widget_class, AdwPreferencesWindow, content_stack);
   gtk_widget_class_bind_template_child_private (widget_class, AdwPreferencesWindow, pages_stack);
   gtk_widget_class_bind_template_child_private (widget_class, AdwPreferencesWindow, search_button);
@@ -632,10 +677,12 @@ adw_preferences_window_class_init (AdwPreferencesWindowClass *klass)
   gtk_widget_class_bind_template_child_private (widget_class, AdwPreferencesWindow, search_results);
   gtk_widget_class_bind_template_child_private (widget_class, AdwPreferencesWindow, search_stack);
   gtk_widget_class_bind_template_child_private (widget_class, AdwPreferencesWindow, title_stack);
-  gtk_widget_class_bind_template_child_private (widget_class, AdwPreferencesWindow, view_switcher_bar);
-  gtk_widget_class_bind_template_child_private (widget_class, AdwPreferencesWindow, view_switcher_title);
-  gtk_widget_class_bind_template_callback (widget_class, subpages_leaflet_child_transition_running_cb);
-  gtk_widget_class_bind_template_callback (widget_class, subpages_leaflet_visible_child_cb);
+  gtk_widget_class_bind_template_child_private (widget_class, AdwPreferencesWindow, view_switcher_stack);
+  gtk_widget_class_bind_template_child_private (widget_class, AdwPreferencesWindow, view_switcher);
+  gtk_widget_class_bind_template_child_private (widget_class, AdwPreferencesWindow, title);
+  gtk_widget_class_bind_template_child_private (widget_class, AdwPreferencesWindow, breakpoint);
+  gtk_widget_class_bind_template_callback (widget_class, try_remove_legacy_subpages);
+  gtk_widget_class_bind_template_callback (widget_class, update_view_switcher);
   gtk_widget_class_bind_template_callback (widget_class, title_stack_notify_transition_running_cb);
   gtk_widget_class_bind_template_callback (widget_class, title_stack_notify_visible_child_cb);
   gtk_widget_class_bind_template_callback (widget_class, notify_visible_page_cb);
@@ -671,7 +718,7 @@ adw_preferences_window_init (AdwPreferencesWindow *self)
   gtk_widget_init_template (GTK_WIDGET (self));
 
   priv->filter = GTK_FILTER (gtk_custom_filter_new ((GtkCustomFilterFunc) filter_search_results, self, NULL));
-  expr = gtk_property_expression_new (GTK_TYPE_STACK_PAGE, NULL, "visible");
+  expr = gtk_property_expression_new (ADW_TYPE_VIEW_STACK_PAGE, NULL, "visible");
 
   model = G_LIST_MODEL (adw_view_stack_get_pages (priv->pages_stack));
   model = G_LIST_MODEL (gtk_filter_list_model_new (model, GTK_FILTER (gtk_bool_filter_new (expr))));
@@ -715,6 +762,8 @@ adw_preferences_window_buildable_init (GtkBuildableIface *iface)
  * Creates a new `AdwPreferencesWindow`.
  *
  * Returns: the newly created `AdwPreferencesWindow`
+ *
+ * Deprecated: 1.6: Use [class@PreferencesDialog].
  */
 GtkWidget *
 adw_preferences_window_new (void)
@@ -728,6 +777,8 @@ adw_preferences_window_new (void)
  * @page: the page to add
  *
  * Adds a preferences page to @self.
+ *
+ * Deprecated: 1.6: Use [class@PreferencesDialog].
  */
 void
 adw_preferences_window_add (AdwPreferencesWindow *self,
@@ -747,6 +798,9 @@ adw_preferences_window_add (AdwPreferencesWindow *self,
   g_object_bind_property (page, "title", stack_page, "title", G_BINDING_SYNC_CREATE);
   g_object_bind_property (page, "use-underline", stack_page, "use-underline", G_BINDING_SYNC_CREATE);
   g_object_bind_property (page, "name", stack_page, "name", G_BINDING_SYNC_CREATE);
+
+  priv->n_pages++;
+  update_view_switcher (self);
 }
 
 /**
@@ -755,6 +809,8 @@ adw_preferences_window_add (AdwPreferencesWindow *self,
  * @page: the page to remove
  *
  * Removes a page from @self.
+ *
+ * Deprecated: 1.6: Use [class@PreferencesDialog].
  */
 void
 adw_preferences_window_remove (AdwPreferencesWindow *self,
@@ -771,6 +827,9 @@ adw_preferences_window_remove (AdwPreferencesWindow *self,
     adw_view_stack_remove (priv->pages_stack, GTK_WIDGET (page));
   else
     ADW_CRITICAL_CANNOT_REMOVE_CHILD (self, page);
+
+  priv->n_pages--;
+  update_view_switcher (self);
 }
 
 /**
@@ -780,6 +839,8 @@ adw_preferences_window_remove (AdwPreferencesWindow *self,
  * Gets the currently visible page of @self.
  *
  * Returns: (transfer none) (nullable): the visible page
+ *
+ * Deprecated: 1.6: Use [class@PreferencesDialog].
  */
 AdwPreferencesPage *
 adw_preferences_window_get_visible_page (AdwPreferencesWindow *self)
@@ -799,6 +860,8 @@ adw_preferences_window_get_visible_page (AdwPreferencesWindow *self)
  * @page: a page of @self
  *
  * Makes @page the visible page of @self.
+ *
+ * Deprecated: 1.6: Use [class@PreferencesDialog].
  */
 void
 adw_preferences_window_set_visible_page (AdwPreferencesWindow *self,
@@ -821,6 +884,8 @@ adw_preferences_window_set_visible_page (AdwPreferencesWindow *self,
  * Gets the name of currently visible page of @self.
  *
  * Returns: (transfer none) (nullable): the name of the visible page
+ *
+ * Deprecated: 1.6: Use [class@PreferencesDialog].
  */
 const char *
 adw_preferences_window_get_visible_page_name (AdwPreferencesWindow *self)
@@ -841,7 +906,9 @@ adw_preferences_window_get_visible_page_name (AdwPreferencesWindow *self)
  *
  * Makes the page with the given name visible.
  *
- * See [property@ViewStack:visible-child].
+ * See [property@PreferencesWindow:visible-page].
+ *
+ * Deprecated: 1.6: Use [class@PreferencesDialog].
  */
 void
 adw_preferences_window_set_visible_page_name (AdwPreferencesWindow *self,
@@ -857,12 +924,14 @@ adw_preferences_window_set_visible_page_name (AdwPreferencesWindow *self,
 }
 
 /**
- * adw_preferences_window_get_search_enabled: (attributes org.gtk.Method.get_property=search-enabled)
+ * adw_preferences_window_get_search_enabled:
  * @self: a preferences window
  *
  * Gets whether search is enabled for @self.
  *
  * Returns: whether search is enabled for @self.
+ *
+ * Deprecated: 1.6: Use [class@PreferencesDialog].
  */
 gboolean
 adw_preferences_window_get_search_enabled (AdwPreferencesWindow *self)
@@ -877,11 +946,13 @@ adw_preferences_window_get_search_enabled (AdwPreferencesWindow *self)
 }
 
 /**
- * adw_preferences_window_set_search_enabled: (attributes org.gtk.Method.set_property=search-enabled)
+ * adw_preferences_window_set_search_enabled:
  * @self: a preferences window
  * @search_enabled: whether search is enabled
  *
  * Sets whether search is enabled for @self.
+ *
+ * Deprecated: 1.6: Use [class@PreferencesDialog].
  */
 void
 adw_preferences_window_set_search_enabled (AdwPreferencesWindow *self,
@@ -911,7 +982,7 @@ adw_preferences_window_set_search_enabled (AdwPreferencesWindow *self,
 }
 
 /**
- * adw_preferences_window_set_can_navigate_back: (attributes org.gtk.Method.set_property=can-navigate-back)
+ * adw_preferences_window_set_can_navigate_back:
  * @self: a preferences window
  * @can_navigate_back: the new value
  *
@@ -927,6 +998,10 @@ adw_preferences_window_set_search_enabled (AdwPreferencesWindow *self,
  * <kbd>Alt</kbd>+<kbd>←</kbd> shortcut.
  *
  * For right-to-left locales, gestures and shortcuts are reversed.
+ *
+ * Has no effect for subpages added with [method@PreferencesWindow.push_subpage].
+ *
+ * Deprecated: 1.4: Use [method@NavigationPage.set_can_pop] instead.
  */
 void
 adw_preferences_window_set_can_navigate_back (AdwPreferencesWindow *self,
@@ -949,12 +1024,14 @@ adw_preferences_window_set_can_navigate_back (AdwPreferencesWindow *self,
 }
 
 /**
- * adw_preferences_window_get_can_navigate_back: (attributes org.gtk.Method.get_property=can-navigate-back)
+ * adw_preferences_window_get_can_navigate_back:
  * @self: a preferences window
  *
  * Gets whether gestures and shortcuts for closing subpages are enabled.
  *
  * Returns: whether gestures and shortcuts are enabled.
+ *
+ * Deprecated: 1.4: Use [method@NavigationPage.get_can_pop] instead.
  */
 gboolean
 adw_preferences_window_get_can_navigate_back (AdwPreferencesWindow *self)
@@ -977,6 +1054,8 @@ adw_preferences_window_get_can_navigate_back (AdwPreferencesWindow *self)
  *
  * The transition can be cancelled by the user, in which case visible child will
  * change back to the previously visible child.
+ *
+ * Deprecated: 1.4: Use [method@PreferencesWindow.push_subpage] instead.
  */
 void
 adw_preferences_window_present_subpage (AdwPreferencesWindow *self,
@@ -1010,6 +1089,8 @@ adw_preferences_window_present_subpage (AdwPreferencesWindow *self,
  * Closes the current subpage.
  *
  * If there is no presented subpage, this does nothing.
+ *
+ * Deprecated: 1.4: Use [method@PreferencesWindow.pop_subpage] instead.
  */
 void
 adw_preferences_window_close_subpage (AdwPreferencesWindow *self)
@@ -1023,7 +1104,56 @@ adw_preferences_window_close_subpage (AdwPreferencesWindow *self)
   if (priv->subpage == NULL)
     return;
 
-  adw_leaflet_set_visible_child (priv->subpages_leaflet, priv->preferences);
+  adw_leaflet_set_visible_child (priv->subpages_leaflet, GTK_WIDGET (priv->subpages_nav_view));
+}
+
+/**
+ * adw_preferences_window_push_subpage:
+ * @self: a preferences window
+ * @page: the subpage
+ *
+ * Pushes @page onto the subpage stack of @self.
+ *
+ * The page will be automatically removed when popped.
+ *
+ * Since: 1.4
+ * Deprecated: 1.6: Use [class@PreferencesDialog].
+ */
+void
+adw_preferences_window_push_subpage (AdwPreferencesWindow *self,
+                                     AdwNavigationPage    *page)
+{
+  AdwPreferencesWindowPrivate *priv;
+
+  g_return_if_fail (ADW_IS_PREFERENCES_WINDOW (self));
+  g_return_if_fail (ADW_IS_NAVIGATION_PAGE (page));
+
+  priv = adw_preferences_window_get_instance_private (self);
+
+  adw_navigation_view_push (priv->subpages_nav_view, page);
+}
+
+/**
+ * adw_preferences_window_pop_subpage:
+ * @self: a preferences window
+ *
+ * Pop the visible page from the subpage stack of @self.
+ *
+ * Returns: `TRUE` if a page has been popped
+ *
+ * Since: 1.4
+ * Deprecated: 1.6: Use [class@PreferencesDialog].
+ */
+gboolean
+adw_preferences_window_pop_subpage (AdwPreferencesWindow *self)
+{
+  AdwPreferencesWindowPrivate *priv;
+
+  g_return_val_if_fail (ADW_IS_PREFERENCES_WINDOW (self), FALSE);
+
+  priv = adw_preferences_window_get_instance_private (self);
+
+  return adw_navigation_view_pop (priv->subpages_nav_view);
 }
 
 /**
@@ -1034,6 +1164,8 @@ adw_preferences_window_close_subpage (AdwPreferencesWindow *self)
  * Displays @toast.
  *
  * See [method@ToastOverlay.add_toast].
+ *
+ * Deprecated: 1.6: Use [class@PreferencesDialog].
  */
 void
 adw_preferences_window_add_toast (AdwPreferencesWindow *self,
@@ -1047,4 +1179,3 @@ adw_preferences_window_add_toast (AdwPreferencesWindow *self,
 
   adw_toast_overlay_add_toast (priv->toast_overlay, toast);
 }
-
