@@ -17,6 +17,9 @@
 
 #include "adw-widget-utils-private.h"
 
+#include "adw-bottom-sheet-private.h"
+#include "adw-floating-sheet-private.h"
+
 typedef struct _CompareInfo CompareInfo;
 
 enum Axis {
@@ -477,6 +480,15 @@ adw_widget_grab_focus_child (GtkWidget *widget)
   return FALSE;
 }
 
+gboolean
+adw_widget_grab_focus_child_or_self (GtkWidget *widget)
+{
+  if (adw_widget_grab_focus_child (widget))
+    return TRUE;
+
+  return adw_widget_grab_focus_self (widget);
+}
+
 void
 adw_widget_compute_expand (GtkWidget *widget,
                            gboolean  *hexpand_p,
@@ -546,6 +558,14 @@ adw_widget_get_request_mode (GtkWidget *widget)
         GTK_SIZE_REQUEST_HEIGHT_FOR_WIDTH;
 }
 
+gboolean
+adw_widget_contains_passthrough (GtkWidget *widget,
+                                 double     x,
+                                 double     y)
+{
+  return FALSE;
+}
+
 /* FIXME: Replace this with public color API and make public */
 gboolean
 adw_widget_lookup_color (GtkWidget  *widget,
@@ -559,3 +579,69 @@ G_GNUC_BEGIN_IGNORE_DEPRECATIONS
 G_GNUC_END_IGNORE_DEPRECATIONS
 }
 
+GtkWidget *
+adw_widget_get_ancestor (GtkWidget *widget,
+                         GType      widget_type,
+                         gboolean   same_native,
+                         gboolean   same_sheet)
+{
+  GtkWidget *prev_widget = NULL;
+
+  while (widget && !g_type_is_a (G_OBJECT_TYPE (widget), widget_type)) {
+    if (same_native && GTK_IS_NATIVE (widget))
+      return NULL;
+
+    if (same_sheet) {
+      if (ADW_IS_FLOATING_SHEET (widget)) {
+        if (prev_widget == adw_floating_sheet_get_sheet_bin (ADW_FLOATING_SHEET (widget)))
+          return NULL;
+      }
+
+      if (ADW_IS_BOTTOM_SHEET (widget)) {
+        if (prev_widget == adw_bottom_sheet_get_sheet_bin (ADW_BOTTOM_SHEET (widget)))
+          return NULL;
+      }
+    }
+
+    prev_widget = widget;
+    widget = gtk_widget_get_parent (widget);
+  }
+
+  return widget;
+}
+
+gboolean
+adw_decoration_layout_prefers_start (const char *layout)
+{
+  int counts[2];
+  char **sides;
+  int i;
+
+  if (!layout || !strchr (layout, ':'))
+    return FALSE;
+
+  sides = g_strsplit (layout, ":", 2);
+
+  for (i = 0; i < 2; i++) {
+    char **elements;
+    int j;
+
+    counts[i] = 0;
+
+    if (sides[i] == NULL)
+      continue;
+
+    elements = g_strsplit (sides[i], ",", -1);
+
+    for (j = 0; elements[j]; j++) {
+      if (!g_strcmp0 (elements[j], "close"))
+        counts[i]++;
+    }
+
+    g_strfreev (elements);
+  }
+
+  g_strfreev (sides);
+
+  return counts[0] > counts[1];
+}
